from functools import lru_cache
from importlib import import_module
from typing import Any, Callable, Union


Executable = Union[str, Callable[..., Any]]


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError('Module "%s" does not define a "%s" attribute/class' % (
            module_path, class_name)
        ) from err


cached_import_string = lru_cache(maxsize=None)(import_string)


def normalize_executable(executable: Executable) -> Executable:
    if isinstance(executable, str):
        return cached_import_string(executable)

    return executable
